package org.osmf.smpte.tt.media
{
	import org.osmf.elements.ParallelElement;
	import org.osmf.media.MediaElement;
	import org.osmf.metadata.TimelineMetadata;
	import org.osmf.net.StreamingURLResource;
	import org.osmf.smpte.tt.captions.CaptionElement;
	import org.osmf.smpte.tt.captions.CaptioningDocument;
	import org.osmf.smpte.tt.captions.TimedTextElement;
	import org.osmf.smpte.tt.events.ParseEvent;
	import org.osmf.smpte.tt.loader.SMPTETTLoadTrait;
	
	public class SMPTETTProxyElementAsync extends SMPTETTProxyElement
	{		
		public function SMPTETTProxyElementAsync(proxiedElement:MediaElement=null, continueLoadOnFailure:Boolean=true)
		{
			super(proxiedElement, continueLoadOnFailure);
		}
		
		override public function set proxiedElement(value:MediaElement):void
		{
			if (value != null)
			{
				super.proxiedElement = value;
			}
		}
		
		override protected function addToParallelElement(parallelElement:ParallelElement, p_captioningMediaElement:CaptioningMediaElement):void
		{
			/*
			Ensure that the value that we are adding is not already part of the element, 
			otherwise you will get an RTE. 
			
			This was a more relevant check it the first iteration, when all of the captions
			were added at each ParseEvent.PARTIAL event. 
			
			Subsequently, it *should* only be adding the new elements, but the check still seems worthwhile
			*/
			if (parallelElement.getChildIndex(p_captioningMediaElement) == -1)
			{
				parallelElement.addChild(p_captioningMediaElement);
			}
		}
		
		/*
		Initially, CaptioningMediaElement could return a null RegionLayoutTargetSprite.
		It has been fixed, but a null value is still possible
		*/
		override protected function regionSpriteAddMediaElement(document:CaptioningDocument, i:uint, p_captioningMediaElement:CaptioningMediaElement):void
		{
			var region:RegionLayoutTargetSprite =
				p_captioningMediaElement.addRegion(document.captionRegions[i]);				
			if (region)
			{
				region.mediaElement = mediaElement;
			}
			else
			{
				trace ("regionSpriteAddMediaElement failed " + i);
			}
		}
		
		
		
		override protected function addLoadTraitListeners(trait:SMPTETTLoadTrait):void
		{
			super.addLoadTraitListeners(trait);
			trait.addEventListener(ParseEvent.BEGIN, onParseEvent, false, 0, true);
			trait.addEventListener(ParseEvent.PROGRESS, onParseEvent, false, 0, true);
			trait.addEventListener(ParseEvent.PARTIAL, onParseEvent, false, 0, true);
			trait.addEventListener(ParseEvent.COMPLETE, onParseEvent, false, 0, true);
		}
		
		protected function onParseEvent(e:ParseEvent):void
		{
			switch(e.type)
			{
				case ParseEvent.PARTIAL:
					onPartialLoad(e);
					break;
			}
			dispatchEvent(e);
		}

		protected function onPartialLoad(e:ParseEvent):void
		{
			addMetaDataTimelineMarkers();
			
			//Make sure on partial loading of captions, we quickly get to the nearest one
			if (_mediaPlayer)
			{	
				if (_wasPlaying)
				{
					_mediaPlayer.play();
					_wasPlaying = false;
				}
				else if (_wasPaused && _mediaPlayer.playing)
				{
					_wasPaused = false;
				}
				showNearestCaption(_mediaPlayer.currentTime);
			}
		}
		
		
		/*
		Instead of building up a full list of captions, the captions are now removed from the array as they are being
		processed through each ParseEvent.PARTIAL event.
		*/
		override protected function addEachCaptionToMetaData(document:CaptioningDocument, SMPTETTMetadata:TimelineMetadata):void
		{
			if (document.captionElements)
			{
				var captionElements:Vector.<CaptionElement> = document.captionElements;
				var c:CaptionElement;
				var i:uint;
				var startAt:int = Math.max(0,SMPTETTMetadata.numMarkers-1);
				for (i=startAt; i<document.numTimelineMarkerEvents; i++)
				{
					c = captionElements[i];
					if (c)
					{												
						// c.end returns a very large value if the captioning sample XML does not have an 'end' attribute.
						// So the subclipping logic won't work. The Akamai captioning sample is not per SMPTE spec it appears.
						// By adding the end or duration attribute, the following can be uncommented.
						
						// Check for sub-clipping
						if (mediaElement != null && mediaElement.resource is StreamingURLResource)
						{
							var res:StreamingURLResource = StreamingURLResource(mediaElement.resource);
							var clipStart:Number = res.clipStartTime;
							var clipEnd:Number = res.clipEndTime;
							if (!isNaN(clipStart) ||
							      !isNaN(clipEnd))
							{
								// Captions with begin and end times prior to clipStart or begin and end times following clipEnd
								// are excluded from the timelineMetadata and do not display.
								if (!includeCaptionInSubClip(c, clipStart, clipEnd))
								{
									continue;
								}
								
								var adjustedCaption:CaptionElement = new CaptionElement(captionStartTime(c, clipStart), 
																						  captionEndTime(c, clipStart, clipEnd), 
																					        c.id);
								if (!cloneCaptionProperties(c, adjustedCaption, res))
								{
									// continue for now.
									continue;
								}
								c = adjustedCaption;
								
								function includeCaptionInSubClip(caption:CaptionElement, clipStart:Number=NaN, clipEnd:Number=NaN):Boolean
								{
									var included:Boolean = true;
									
									if ((!isNaN(clipStart) &&
										  caption.begin <= clipStart &&
										    caption.end <= clipStart) ||
										(!isNaN(clipEnd) &&
											caption.begin >= clipEnd &&
											  caption.end >= clipEnd))
									{
										included = false;
									}
									return included;
								}
								
								function captionStartTime(caption:CaptionElement, clipStart:Number=NaN):Number
								{
									// return a modified caption begin time based on clipStart.
									return (isNaN(clipStart)) ? caption.begin : ((caption.begin <= clipStart) ? clipStart : caption.begin - clipStart);
								}
								
								function captionEndTime(caption:CaptionElement, clipStart:Number=NaN, clipEnd:Number=NaN):Number
								{
									// return a modified caption end time based on clipStart and clipEnd.
									return (isNaN(clipEnd)) ? ((isNaN(clipStart) ? caption.end : caption.end - clipStart)) : 
										      ((caption.end <= clipEnd) ? (caption.end - clipStart) : (clipEnd - clipStart));
								}
							}
						}
						
						SMPTETTMetadata.addMarker(c);
						if (SMPTETTProxyElement.DEBUG) addCaptionElementToDisplay(c);
					}
				}
			}
		}
		
		// Clones all writable properties of a CaptionElement, including children.
		// Useful for creating caption elements with modified begin and end times.
		private function cloneCaptionProperties(predecessor:CaptionElement, successor:CaptionElement, resource:StreamingURLResource):Boolean
		{			
			if (predecessor == null || 
				successor == null)
			{
				return false;
			}
			successor.captionElementType = predecessor.captionElementType;
			successor.style = predecessor.style.clone();
			successor.content = predecessor.content;
			successor.regionId = predecessor.regionId;
			successor.index = predecessor.index;
			successor.parentElement = predecessor.parentElement;
			
			// Iterate through the children and assign required properties.
			for each (var child:TimedTextElement in predecessor.children)
			{
				var successorChild:CaptionElement = new CaptionElement(child.time - resource.clipStartTime, child.end - resource.clipStartTime, child.id);
				cloneCaptionProperties(CaptionElement(child), successorChild, resource);
				successor.children.push(successorChild);
			}
			return true;
		}
	}
}